import sys
import warnings

import matplotlib.pyplot as plt
import numpy as np
from bokeh.plotting import figure, output_file, show


class PressureMatrix:
    r"""This class calculates the pressure matrix of the object with the given parameters.

    It is supposed to be an attribute of a bearing element,
    but can work on its on to provide graphics for the user.

    Parameters
    ----------
    Grid related
    ^^^^^^^^^^^^
    Describes the discretization of the problem
    nz: int
        Number of points along the Z direction (direction of flow).
    ntheta: int
        Number of points along the direction theta. NOTE: ntheta must be odd.
    nradius: int
        Number of points along the direction r.
    length: float
        Length in the Z direction (m).

    Operation conditions
    ^^^^^^^^^^^^^^^^^^^^
    Describes the operation conditions.
    omega: float
        Rotation of the rotor (rad/s).
    p_in: float
        Input Pressure (Pa).
    p_out: float
        Output Pressure (Pa).
    load: float
        Load applied to the rotor (N).

    Geometric data of the problem
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    Describes the geometric data of the problem.
    radius_rotor: float
        Rotor radius (m).
    radius_stator: float
        Stator Radius (m).
    eccentricity: float
        Eccentricity (m) is the euclidean distance between rotor and stator centers.
        The center of the stator is in position (0,0).
    beta: float
        Angle between the origin and the eccentricity (rad).

    Fluid characteristics
    ^^^^^^^^^^^^^^^^^^^^^
    Describes the fluid characteristics.
    visc: float
        Viscosity (Pa.s).
    rho: float
        Fluid density(Kg/m^3).

    Returns
    -------
    An object containing the pressure matrix and its data.

    Attributes
    ----------
    ltheta: float
        Length in the theta direction (rad).
    dz: float
        Range size in the Z direction.
    dtheta: float
        Range size in the theta direction.
    ntotal: int
        Number of nodes in the grid., ntheta, nradius, n_interv_z, n_interv_theta,
    n_interv_z: int
        Number of intervals on Z.
    n_interv_theta: int
      Number of intervals on theta.
    n_interv_radius: int
        Number of intervals on r.
    p_mat_analytical : array of shape (nz, ntheta)
        The analytical pressure matrix.
    p_mat_numerical : array of shape (nz, ntheta)
        The numerical pressure matrix.
    xe: float
        Eccentricity (m) (distance between rotor and stator centers) on the x-axis.
        It is the position of the center of the stator.
        The center of the rotor is in position (0,0).
    ye: float
        Eccentricity (m) (distance between rotor and stator centers) on the y-axis.
        It is the position of the center of the stator.
        The center of the rotor is in position (0,0).
    re : array of shape (nz, ntheta)
        The external radius in each position of the grid.
    ri : array of shape (nz, ntheta)
        The internal radius in each position of the grid.
    z : array of shape (1, nz)
        z along the object. It goes from 0 to lb.
    xre : array of shape (nz, ntheta)
        x value of the external radius.
    xri : array of shape (nz, ntheta)
        x value of the internal radius.
    yre : array of shape (nz, ntheta)
        y value of the external radius.
    yri : array of shape (nz, ntheta)
        y value of the internal radius.
    eccentricity : float
        distance between the center of the rotor and the stator.
    difference_between_radius: float
        distance between the radius of the stator and the radius of the rotor.
    eccentricity_ratio: float
        eccentricity/difference_between_radius
    bearing_type: str
        type of structure. 'short_bearing': short; 'long_bearing': long;
        'medium_size': in between short and long.
        if length/radius_stator <= 1/8 it is short.
        if length/radius_stator > 4 it is long.
    plot_counter: int
        for each plot, it increments this counter, in order to plot
        figures with unique IDs.
    radial_clearance: float
        Difference between both stator and rotor radius, regardless of eccentricity.

    Examples
    --------
    >>> from ross.fluid_flow import fluid_flow as flow
    >>> import numpy as np
    >>> nz = 8
    >>> ntheta = 64
    >>> nradius = 11
    >>> length = 0.01
    >>> omega = 100.*2*np.pi/60
    >>> p_in = 1.
    >>> p_out = 1.
    >>> radius_rotor = 0.08
    >>> radius_stator = 0.1
    >>> visc = 0.015
    >>> rho = 860.
    >>> eccentricity = 0.001
    >>> beta =  np.pi
    >>> my_pressure_matrix = flow.PressureMatrix(nz, ntheta, nradius, length,
    ...                                          omega, p_in, p_out, radius_rotor,
    ...                                          radius_stator, visc, rho, beta=beta, eccentricity=eccentricity)
    >>> my_pressure_matrix.calculate_pressure_matrix_analytical() # doctest: +ELLIPSIS
    array([[-0.00000...
    >>> my_pressure_matrix.calculate_pressure_matrix_numerical() # doctest: +ELLIPSIS
    array([[1...
    >>> my_pressure_matrix.plot_eccentricity()
    >>> my_pressure_matrix.plot_pressure_z()
    >>> my_pressure_matrix.plot_shape()
    >>> my_pressure_matrix.plot_pressure_theta(z=int(nz/2))
    >>> my_pressure_matrix.matplot_pressure_theta_cylindrical(z=int(nz/2),
    ... show_immediately=False)
    """

    def __init__(
        self,
        nz,
        ntheta,
        nradius,
        length,
        omega,
        p_in,
        p_out,
        radius_rotor,
        radius_stator,
        visc,
        rho,
        beta=2.356,
        eccentricity=None,
        load=None,
    ):
        if load is None and eccentricity is None:
            sys.exit("Either load or eccentricity must be given.")
        self.nz = nz
        self.ntheta = ntheta
        self.nradius = nradius
        self.n_interv_z = nz - 1
        self.n_interv_theta = ntheta - 1
        self.n_interv_radius = nradius - 1
        self.length = length
        self.ltheta = 2.0 * np.pi
        self.dz = length / self.n_interv_z
        self.dtheta = self.ltheta / self.n_interv_theta
        self.ntotal = self.nz * self.ntheta
        self.omega = omega
        self.p_in = p_in
        self.p_out = p_out
        self.radius_rotor = radius_rotor
        self.radius_stator = radius_stator
        self.visc = visc
        self.rho = rho
        self.radial_clearance = self.radius_stator - self.radius_rotor
        self.difference_between_radius = radius_stator - radius_rotor
        self.bearing_type = ""
        if self.length / self.radius_stator <= 1 / 8:
            self.bearing_type = "short_bearing"
        elif self.length / self.radius_stator > 4:
            self.bearing_type = "long_bearing"
        else:
            self.bearing_type = "medium_size"
        self.eccentricity = eccentricity
        self.beta = beta
        self.eccentricity_ratio = None
        self.load = load
        if self.eccentricity is None:
            self.eccentricity = (
                self.calculate_eccentricity_ratio() * self.difference_between_radius
            )
        self.eccentricity_ratio = self.eccentricity / self.difference_between_radius
        if self.load is None:
            self.load = self.get_rotor_load()
        self.xe = self.eccentricity * np.cos(beta)
        self.ye = self.eccentricity * np.sin(beta)
        self.re = np.zeros([self.nz, self.ntheta])
        self.ri = np.zeros([self.nz, self.ntheta])
        self.z = np.zeros([1, self.nz])
        self.xre = np.zeros([self.nz, self.ntheta])
        self.xri = np.zeros([self.nz, self.ntheta])
        self.yre = np.zeros([self.nz, self.ntheta])
        self.yri = np.zeros([self.nz, self.ntheta])
        self.p_mat_analytical = np.zeros([self.nz, self.ntheta])
        self.c1 = np.zeros([self.nz, self.ntheta])
        self.c2 = np.zeros([self.nz, self.ntheta])
        self.c0w = np.zeros([self.nz, self.ntheta])
        self.M = np.zeros([self.ntotal, self.ntotal])
        self.f = np.zeros([self.ntotal, 1])
        self.P = np.zeros([self.ntotal, 1])
        self.p_mat_numerical = np.zeros([self.nz, self.ntheta])
        self.plot_counter = 0
        self.calculate_coefficients()
        self.mounting_matrix()
        self.resolves_matrix()
        self.pressure_matrix_available = False

    def calculate_pressure_matrix_analytical(self):
        """This function calculates the analytical pressure matrix
        """
        if self.bearing_type == "short_bearing":
            for i in range(self.nz):
                for j in range(self.ntheta):
                    # fmt: off
                    self.p_mat_analytical[i][j] = (((-3*self.visc*self.omega)/self.difference_between_radius**2) *
                                                   ((i * self.dz - (self.length / 2)) ** 2 - (self.length ** 2) / 4) *
                                                   (self.eccentricity_ratio * np.sin(j * self.dtheta)) /
                                                   (1 + self.eccentricity_ratio * np.cos(j * self.dtheta))**3)
                    # fmt: on
            self.pressure_matrix_available = True
        return self.p_mat_analytical

    def calculate_coefficients(self):
        """This function calculates the constants that form the Poisson equation
        of the discrete pressure (central differences in the second
        derivatives). It is executed when the class is instantiated.
        """

        for i in range(self.nz):
            zno = i * self.dz
            self.z[0][i] = zno
            plot_eccentricity_error = False
            position = -1
            for j in range(self.ntheta):
                # fmt: off
                gama = j * self.dtheta + self.beta
                [radius_external, self.xre[i][j], self.yre[i][j]] =\
                    self.external_radius_function(gama)
                [radius_internal, self.xri[i][j], self.yri[i][j]] =\
                    self.internal_radius_function(zno, gama)
                self.re[i][j] = radius_external
                self.ri[i][j] = radius_internal

                w = self.omega * self.ri[i][j]

                k = (self.re[i][j] ** 2 * (np.log(self.re[i][j]) - 1 / 2) - self.ri[i][j] ** 2 *
                     (np.log(self.ri[i][j]) - 1 / 2)) / (self.ri[i][j] ** 2 - self.re[i][j] ** 2)

                self.c1[i][j] = (1 / (4 * self.visc)) * ((self.re[i][j] ** 2 * np.log(self.re[i][j]) -
                                                          self.ri[i][j] ** 2 * np.log(self.ri[i][j]) +
                                                          (self.re[i][j] ** 2 - self.ri[i][j] ** 2) *
                                                          (k - 1)) - 2 * self.re[i][j] ** 2 *
                                                         ((np.log(self.re[i][j]) + k - 1 / 2) * np.log(self.re[i][j] /
                                                                                                       self.ri[i][j])))

                self.c2[i][j] = (- self.ri[i][j] ** 2) / (8 * self.visc) *\
                                ((self.re[i][j] ** 2 - self.ri[i][j] ** 2 -
                                  (self.re[i][j] ** 4 - self.ri[i][j] ** 4) /
                                  (2 * self.ri[i][j] ** 2)) +
                                 ((self.re[i][j] ** 2 - self.ri[i][j] ** 2) /
                                  (self.ri[i][j] ** 2 *
                                   np.log(self.re[i][j] / self.ri[i][j]))) *
                                 (self.re[i][j] ** 2 * np.log(self.re[i][j] / self.ri[i][j]) -
                                  (self.re[i][j] ** 2 - self.ri[i][j] ** 2) / 2))

                self.c0w[i][j] = (- w * self.ri[i][j] *
                                  (np.log(self.re[i][j] / self.ri[i][j]) *
                                  (1 + (self.ri[i][j] ** 2) / (self.re[i][j] ** 2 - self.ri[i][j] ** 2)) - 1 / 2))
                # fmt: on
                if not plot_eccentricity_error:
                    if abs(self.xri[i][j]) > abs(self.xre[i][j]) or abs(
                        self.yri[i][j]
                    ) > abs(self.yre[i][j]):
                        plot_eccentricity_error = True
                        position = i
            if plot_eccentricity_error:
                self.plot_eccentricity(position)
                sys.exit(
                    "Error: The given parameters create a rotor that is not inside the stator. "
                    "Check the plotted figure and fix accordingly."
                )

    def mounting_matrix(self):
        """This function assembles the matrix M and the independent vector f
        """

        # fmt: off
        count = 0
        for x in range(self.ntheta):
            self.M[count][count] = 1
            self.f[count][0] = self.p_in
            count = count + self.nz - 1
            self.M[count][count] = 1
            self.f[count][0] = self.p_out
            count = count + 1
        count = 0
        for x in range(self.nz - 2):
            self.M[self.ntotal - self.nz + 1 + count][1 + count] = 1
            self.M[self.ntotal - self.nz + 1 + count][self.ntotal - self.nz + 1 + count] = -1
            count = count + 1
        count = 1
        j = 0
        for i in range(1, self.nz - 1):
            a = (1/self.dtheta**2)*(self.c1[i][self.ntheta-1])
            self.M[count][self.ntotal - 2 * self.nz + count] = a
            b = (1 / self.dz ** 2) * (self.c2[i - 1, j])
            self.M[count][count - 1] = b
            c = -((1/self.dtheta**2) * ((self.c1[i][j]) + self.c1[i][self.ntheta-1])
                  + (1/self.dz**2) * (self.c2[i][j] + self.c2[i-1][j]))
            self.M[count, count] = c
            d = (1 / self.dz ** 2) * (self.c2[i][j])
            self.M[count][count + 1] = d
            e = (1/self.dtheta**2) * (self.c1[i][j])
            self.M[count][count + self.nz] = e
            count = count + 1
        count = self.nz + 1
        for j in range(1, self.ntheta - 1):
            for i in range(1, self.nz - 1):
                a = (1/self.dtheta**2) * (self.c1[i, j-1])
                self.M[count][count - self.nz] = a
                b = (1 / self.dz ** 2) * (self.c2[i - 1][j])
                self.M[count][count - 1] = b
                c = -((1 / self.dtheta**2) * ((self.c1[i][j]) + self.c1[i][j-1])
                      + (1 / self.dz**2) * (self.c2[i][j] + self.c2[i-1][j]))
                self.M[count, count] = c
                d = (1 / self.dz ** 2) * (self.c2[i][j])
                self.M[count][count + 1] = d
                e = (1 / self.dtheta**2) * (self.c1[i][j])
                self.M[count][count + self.nz] = e
                count = count + 1
            count = count + 2
        count = 1
        for j in range(self.ntheta - 1):
            for i in range(1, self.nz - 1):
                if j == 0:
                    self.f[count][0] = (self.c0w[i][j] - self.c0w[i][self.ntheta - 1]) / self.dtheta
                else:
                    self.f[count][0] = (self.c0w[i, j] - self.c0w[i, j - 1]) / self.dtheta
                count = count + 1
            count = count + 2
        # fmt: on

    def resolves_matrix(self):
        """This function resolves the linear system [M]{P}={f}
        """
        self.P = np.linalg.solve(self.M, self.f)

    def calculate_pressure_matrix_numerical(self):
        """This function calculates the numerical pressure matrix
        """
        for i in range(self.nz):
            for j in range(self.ntheta):
                k = j * self.nz + i
                self.p_mat_numerical[i][j] = self.P[k]
        return self.p_mat_numerical

    def internal_radius_function(self, z, gama):
        """This function calculates the radius of the rotor given the
        radius and the position z.
        Parameters
        ----------
        z: float
            Distance along the z-axis.
        gama: float
            Gama is the distance in the theta-axis. It should range from 0 to 2*np.pi.
        Returns
        -------
        radius_internal: float
            The size of the internal radius at that point.
        xri: float
            The position x of the returned internal radius.
        yri: float
            The position y of the returned internal radius.
        """
        radius_internal = self.radius_rotor
        xri = radius_internal * np.cos(gama)
        yri = radius_internal * np.sin(gama)
        return radius_internal, xri, yri

    def external_radius_function(self, gama):
        """This function calculates the radius of the stator.
        Parameters
        ----------
        gama: float
            Gama is the distance in the theta-axis. It should range from 0 to 2*np.pi.
        Returns
        -------
        radius_external: float
            The size of the external radius at that point.
        xre: float
            The position x of the returned external radius.
        yre: float
            The position y of the returned external radius.
        """
        alpha = np.absolute(self.beta - gama)
        radius_external = np.sqrt(
            self.radius_stator ** 2 - (self.eccentricity * np.sin(alpha)) ** 2
        ) + self.eccentricity * np.cos(alpha)
        xre = radius_external * np.cos(gama)
        yre = radius_external * np.sin(gama)

        return radius_external, xre, yre

    def get_rotor_load(self):
        """Returns the load applied to the rotor, based on the eccentricity ratio.
        Suitable only for short bearings.
        Returns
        -------
        float
            Load applied to the rotor.
        """
        if not self.bearing_type == "short_bearing":
            warnings.warn(
                "Function get_rotor_load suitable only for short bearings. "
                "The ratio between the bearing length and its radius should be less or "
                "equal to 0.125. Currently we have "
                + str(self.length / self.radius_stator)
                + "."
            )
        return (
            (
                np.pi
                * self.radius_stator
                * 2
                * self.omega
                * self.visc
                * (self.length ** 3)
                * self.eccentricity_ratio
            )
            / (
                8
                * (self.radial_clearance ** 2)
                * ((1 - self.eccentricity_ratio ** 2) ** 2)
            )
        ) * (np.sqrt((16 / (np.pi ** 2) - 1) * self.eccentricity_ratio ** 2 + 1))

    def modified_sommerfeld_number(self):
        """Returns the modified sommerfeld number.
        Returns
        -------
        float
            The modified sommerfeld number.
        """
        return (
            self.radius_stator * 2 * self.omega * self.visc * (self.length ** 3)
        ) / (8 * self.load * (self.radial_clearance ** 2))

    def sommerfeld_number(self):
        """Returns the sommerfeld number.
        Returns
        -------
        float
            The sommerfeld number.
        """
        modified_s = self.modified_sommerfeld_number()
        return (modified_s / np.pi) * (self.radius_stator * 2 / self.length) ** 2

    def calculate_eccentricity_ratio(self):
        """Calculate the eccentricity ratio for a given load using the sommerfeld number.
        Suitable only for short bearings.
        Returns
        -------
        float
            The eccentricity ratio.
        """
        if not self.bearing_type == "short_bearing":
            warnings.warn(
                "Function calculate_eccentricity_ratio suitable only for short bearings. "
                "The ratio between the bearing length and its radius should be less or "
                "equal to 0.125. Currently we have "
                + str(self.length / self.radius_stator)
                + "."
            )
        s = self.modified_sommerfeld_number()
        coefficients = [
            1,
            -4,
            (6 - (s ** 2) * (16 - np.pi ** 2)),
            -(4 + (np.pi ** 2) * (s ** 2)),
            1,
        ]
        roots = np.roots(coefficients)
        for i in range(0, len(roots)):
            if 0 <= roots[i] <= 1:
                return np.sqrt(roots[i])
        sys.exit("Eccentricity ratio could not be calculated.")

    def get_analytical_stiffness_matrix(self):
        """Returns the stiffness matrix calculated analytically.
        Suitable only for short bearings.
        Returns
        -------
        list of floats
            A list of length four including stiffness floats in this order: kxx, kxy, kyx, kyy
        """
        if not self.bearing_type == "short_bearing":
            warnings.warn(
                "Function get_analytical_stiffness_matrix suitable only for short bearings. "
                "The ratio between the bearing length and its radius should be less or "
                "equal to 0.125. Currently we have "
                + str(self.length / self.radius_stator)
                + "."
            )
        # fmt: off
        f = self.get_rotor_load()
        h0 = 1.0/(((np.pi**2)*(1 - self.eccentricity_ratio**2) + 16*self.eccentricity_ratio**2)**1.5)
        a = f/self.radial_clearance
        kxx = a*h0*4*((np.pi**2)*(2 - self.eccentricity_ratio**2) + 16*self.eccentricity_ratio**2)
        kxy = (a*h0*np.pi*((np.pi**2)*(1 - self.eccentricity_ratio**2)**2 -
               16*self.eccentricity_ratio**4) /
               (self.eccentricity_ratio*np.sqrt(1 - self.eccentricity_ratio**2)))
        kyx = (-a*h0*np.pi*((np.pi**2)*(1 - self.eccentricity_ratio**2) *
               (1 + 2*self.eccentricity_ratio**2) +
               (32*self.eccentricity_ratio**2)*(1 + self.eccentricity_ratio**2)) /
               (self.eccentricity_ratio*np.sqrt(1 - self.eccentricity_ratio**2)))
        kyy = (a*h0*4*((np.pi**2)*(1 + 2*self.eccentricity_ratio**2) +
                                  ((32*self.eccentricity_ratio**2) *
                                   (1 + self.eccentricity_ratio**2))/(1 - self.eccentricity_ratio**2)))
        # fmt: on
        return [kxx, kxy, kyx, kyy]

    def get_analytical_damping_matrix(self):
        """Returns the damping matrix calculated analytically.
        Suitable only for short bearings.
        Returns
        -------
        list of floats
            A list of length four including stiffness floats in this order: cxx, cxy, cyx, cyy
        """
        if not self.bearing_type == "short_bearing":
            warnings.warn(
                "Function get_analytical_damping_matrix suitable only for short bearings. "
                "The ratio between the bearing length and its radius should be less or "
                "equal to 0.125. Currently we have "
                + str(self.length / self.radius_stator)
                + "."
            )
        # fmt: off
        f = self.get_rotor_load()
        h0 = 1.0/(((np.pi**2)*(1 - self.eccentricity_ratio**2) + 16*self.eccentricity_ratio**2)**1.5)
        a = f/(self.radial_clearance * self.omega)
        cxx = (a*h0*2*np.pi*np.sqrt(1 - self.eccentricity_ratio**2) *
               ((np.pi**2)*(1 + 2*self.eccentricity_ratio**2) - 16*self.eccentricity_ratio**2)/self.eccentricity_ratio)
        cxy = (-a*h0*8*((np.pi**2)*(1 + 2*self.eccentricity_ratio**2) - 16*self.eccentricity_ratio**2))
        cyx = cxy
        cyy = (a*h0*(2*np.pi*((np.pi**2)*(1 - self.eccentricity_ratio**2)**2 + 48*self.eccentricity_ratio**2)) /
               (self.eccentricity_ratio*np.sqrt(1 - self.eccentricity_ratio**2)))
        # fmt: on
        return [cxx, cxy, cyx, cyy]

    def plot_eccentricity(self, z=0):
        """This function assembles pressure graphic along the z-axis.
        The first few plots are of a different color to indicate where theta begins.
        Parameters
        ----------
        z: int
            The distance in z where to cut and plot.
        """
        output_file("plot_eccentricity.html")
        p = figure(
            title="Cut in plane Z=" + str(z),
            x_axis_label="X axis",
            y_axis_label="Y axis",
        )
        for j in range(0, self.ntheta):
            p.circle(self.xre[z][j], self.yre[z][j], color="red")
            p.circle(self.xri[z][j], self.yri[z][j], color="blue")
            p.circle(0, 0, color="blue")
            p.circle(self.xe, self.ye, color="red")
        p.circle(0, 0, color="black")
        show(p)

    def plot_pressure_z(self, theta=0):
        """This function assembles pressure graphic along the z-axis.
        Parameters
        ----------
        theta: int
            The theta to be considered.
        """
        if not self.pressure_matrix_available:
            sys.exit(
                "Must calculate the pressure matrix."
                "Try calling calculate_pressure_matrix first."
            )
        output_file("plot_pressure_z.html")
        x = []
        y = []
        for i in range(0, self.nz):
            x.append(i * self.dz)
            y.append(self.p_mat_analytical[i][theta])
        p = figure(
            title="Pressure along the Z direction (direction of flow); Theta="
            + str(theta),
            x_axis_label="Points along Z",
        )
        p.line(x, y, legend="Analytical pressure", line_width=2)
        show(p)

    def plot_shape(self, theta=0):
        """This function assembles a graphic representing the geometry of the rotor.
        Parameters
        ----------
        theta: int
            The theta to be considered.
        """
        output_file("plot_shape.html")
        x = np.zeros(self.nz)
        y_re = np.zeros(self.nz)
        y_ri = np.zeros(self.nz)
        for i in range(0, self.nz):
            x[i] = i * self.dz
            y_re[i] = self.re[i][theta]
            y_ri[i] = self.ri[i][theta]
        p = figure(
            title="Shapes of stator and rotor along Z; Theta=" + str(theta),
            x_axis_label="Points along Z",
            y_axis_label="Radial direction",
        )
        p.line(x, y_re, line_width=2, color="red")
        p.line(x, y_ri, line_width=2, color="blue")
        show(p)

    def plot_pressure_theta(self, z=0):
        """This function assembles pressure graphic in the theta direction for a given z.
        Parameters
        ----------
        z: int
            The distance along z-axis to be considered.
        """
        if not self.pressure_matrix_available:
            sys.exit(
                "Must calculate the pressure matrix."
                "Try calling calculate_pressure_matrix first."
            )
        output_file("plot_pressure_theta.html")
        theta_list = []
        for theta in range(0, self.ntheta):
            theta_list.append(theta * self.dtheta)
        p = figure(
            title="Pressure along Theta; Z=" + str(z),
            x_axis_label="Points along Theta",
            y_axis_label="Pressure",
        )
        p.line(theta_list, self.p_mat_analytical[z], line_width=2, color="red")
        p.line(theta_list, self.p_mat_numerical[z], line_width=2, color="blue")
        show(p)

    def matplot_eccentricity(self, z=0, show_immediately=True):
        """This function assembles pressure graphic along the z-axis using matplotlib.
        The first few plots are of a different color to indicate where theta begins.
        Parameters
        ----------
        z: int
            The distance in z where to cut and plot.
        show_immediately: bool
            If True, immediately plots the graphic. Otherwise, the user should call plt.show()
            at some point. It is useful in case the user wants to see one graphic alongside another.
        """
        plt.figure(self.plot_counter)
        self.plot_counter += 1
        for j in range(0, self.ntheta):
            plt.plot(self.xre[z][j], self.yre[z][j], "r.")
            plt.plot(self.xri[z][j], self.yri[z][j], "b.")
            plt.plot(0, 0, "b*")
            plt.plot(self.xe, self.ye, "r*")
            plt.title("Cut in plane Z=" + str(z))
            plt.xlabel("X axis")
            plt.ylabel("Y axis")
            plt.axis("equal")
        plt.show(block=show_immediately)
        if show_immediately:
            plt.close("all")

    def matplot_pressure_z(self, show_immediately=True):
        """This function assembles pressure graphic along the z-axis using matplotlib.
        Parameters
        ----------
        show_immediately: bool
            If True, immediately plots the graphic. Otherwise, the user should call plt.show()
            at some point. It is useful in case the user wants to see one graphic alongside another.
        """
        if not self.pressure_matrix_available:
            sys.exit(
                "Must calculate the pressure matrix."
                "Try calling calculate_pressure_matrix first."
            )
        plt.figure(self.plot_counter)
        self.plot_counter += 1
        for i in range(0, self.nz):
            plt.plot(i * self.dz, self.p_mat_analytical[i][0], "bo")
        plt.title("Pressure along the Z direction (direction of flow); Theta=0")
        plt.xlabel("Points along Z")
        plt.ylabel("Pressure")
        plt.show(block=show_immediately)
        if show_immediately:
            plt.close("all")

    def matplot_shape(self, theta=0, show_immediately=True):
        """This function assembles a graphic representing the geometry of the rotor using matplotlib.
        Parameters
        ----------
        theta: int
            The theta to be considered.
        show_immediately: bool
            If True, immediately plots the graphic. Otherwise, the user should call plt.show()
            at some point. It is useful in case the user wants to see one graphic alongside another.
        """
        plt.figure(self.plot_counter)
        self.plot_counter += 1
        x = np.zeros(self.nz)
        y_re = np.zeros(self.nz)
        y_ri = np.zeros(self.nz)
        for i in range(0, self.nz):
            x[i] = i * self.dz
            y_re[i] = self.re[i][theta]
            y_ri[i] = self.ri[i][theta]
        plt.plot(x, y_re, "r")
        plt.plot(x, y_ri, "b")
        plt.title("Shapes of stator and rotor along Z; Theta=" + str(theta))
        plt.xlabel("Points along Z")
        plt.ylabel("Radial direction")
        plt.show(block=show_immediately)
        if show_immediately:
            plt.close("all")

    def matplot_pressure_theta_cylindrical(self, z=0, show_immediately=True):
        """This function assembles cylindrical pressure graphic in the theta direction for a given z,
        using matplotlib.
        Parameters
        ----------
        z: int
            The distance along z-axis to be considered.
        show_immediately: bool
            If True, immediately plots the graphic. Otherwise, the user should call plt.show()
            at some point. It is useful in case the user wants to see one graphic alongside another.
        """
        if not self.pressure_matrix_available:
            sys.exit(
                "Must calculate the pressure matrix."
                "Try calling calculate_pressure_matrix first."
            )
        r = np.arange(
            0,
            self.radius_stator + 0.0001,
            (self.radius_stator - self.radius_rotor) / self.nradius,
        )
        theta = np.arange(-np.pi * 0.25, 1.75 * np.pi + self.dtheta / 2, self.dtheta)

        pressure_along_theta = np.zeros(self.ntheta)
        for i in range(0, self.ntheta):
            pressure_along_theta[i] = self.p_mat_analytical[0][i]

        min_pressure = np.amin(pressure_along_theta)

        r_matrix, theta_matrix = np.meshgrid(r, theta)
        z_matrix = np.zeros((theta.size, r.size))
        inner_radius_list = np.zeros(self.ntheta)
        pressure_list = np.zeros((theta.size, r.size))
        for i in range(0, theta.size):
            inner_radius = np.sqrt(
                self.xri[z][i] * self.xri[z][i] + self.yri[z][i] * self.yri[z][i]
            )
            inner_radius_list[i] = inner_radius
            for j in range(0, r.size):
                if r_matrix[i][j] < inner_radius:
                    continue
                pressure_list[i][j] = pressure_along_theta[i]
                z_matrix[i][j] = pressure_along_theta[i] - min_pressure + 0.01
        fig, ax = plt.subplots(subplot_kw=dict(projection="polar"))
        self.plot_counter += 1
        ax.contourf(theta_matrix, r_matrix, z_matrix, cmap="coolwarm")
        plt.title("Pressure along Theta; Z=" + str(z))
        plt.show(block=show_immediately)
        if show_immediately:
            plt.close("all")

    def matplot_pressure_theta(self, z=0, show_immediately=True):
        """This function assembles pressure graphic in the theta direction for a given z,
        using matplotlib.
        Parameters
        ----------
        z: int
            The distance along z-axis to be considered.
        show_immediately: bool
            If True, immediately plots the graphic. Otherwise, the user should call plt.show()
            at some point. It is useful in case the user wants to see one graphic alongside another.
        """
        plt.figure(self.plot_counter)
        self.plot_counter += 1
        if not self.pressure_matrix_available:
            sys.exit(
                "Must calculate the pressure matrix."
                "Try calling calculate_pressure_matrix first."
            )
        list_of_thetas = []
        for t in range(0, self.ntheta):
            list_of_thetas.append(t * self.dtheta)
        plt.plot(list_of_thetas, self.p_mat_analytical[z], "b")
        plt.title("Pressure along Theta; Z=" + str(z))
        plt.xlabel("Points along Theta")
        plt.ylabel("Pressure")
        plt.show(block=show_immediately)
        if show_immediately:
            plt.close("all")



import pickle

import bokeh.palettes as bp
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from bokeh.layouts import gridplot
from bokeh.models import ColumnDataSource, ColorBar
from bokeh.plotting import figure, output_file, show
from bokeh.transform import linear_cmap

# set bokeh palette of colors
bokeh_colors = bp.RdGy[11]


class Results(np.ndarray):
    """Class used to store results and provide plots.
    This class subclasses np.ndarray to provide additional info and a plot
    method to the calculated results from Rotor.
    Metadata about the results should be stored on info as a dictionary to be
    used on plot configurations and so on.
    Additional attributes can be passed as a dictionary in new_attributes kwarg.
    """

    def __new__(cls, input_array, new_attributes=None):
        obj = np.asarray(input_array).view(cls)

        for k, v in new_attributes.items():
            setattr(obj, k, v)

        # save new attributes names to create them on array finalize
        obj._new_attributes = new_attributes

        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return

        try:
            for k, v in obj._new_attributes.items():
                setattr(self, k, getattr(obj, k, v))
        except AttributeError:
            return

    def __reduce__(self):

        pickled_state = super().__reduce__()
        new_state = pickled_state[2] + (self._new_attributes,)

        return pickled_state[0], pickled_state[1], new_state

    def __setstate__(self, state):
        self._new_attributes = state[-1]
        for k, v in self._new_attributes.items():
            setattr(self, k, v)
        super().__setstate__(state[0:-1])

    def save(self, file):
        with open(file, mode="wb") as f:
            pickle.dump(self, f)

    def plot(self, *args, **kwargs):
        raise NotImplementedError


class CampbellResults(Results):
    def plot(self, harmonics=[1], wn=False, fig=None, ax=None, **kwargs):
        """Plot campbell results.
        Parameters
        ----------
        harmonics: list, optional
            List withe the harmonics to be plotted.
            The default is to plot 1x.
        fig : matplotlib figure, optional
            Figure to insert axes with log_dec colorbar.
        ax : matplotlib axes, optional
            Axes in which the plot will be drawn.
        Returns
        -------
        ax : matplotlib axes
            Returns the axes object with the plot.
        """

        # results for campbell is an array with [speed_range, wd/log_dec/whirl]

        if fig is None and ax is None:
            fig, ax = plt.subplots()

        wd = self[..., 0]
        if wn is True:
            wn = self[..., 4]
        log_dec = self[..., 1]
        whirl = self[..., 2]
        speed_range = self[..., 3]

        default_values = dict(cmap="RdBu", vmin=0.1, vmax=2.0, s=20, alpha=0.5)
        for k, v in default_values.items():
            kwargs.setdefault(k, v)

        # bokeh plot - output to static HTML file
        output_file("Campbell_diagram.html")

        log_dec_map = log_dec.flatten()

        # bokeh plot - create a new plot
        camp = figure(
            tools="pan, box_zoom, wheel_zoom, reset, save",
            width=900,
            height=500,
            title="Campbell Diagram",
            x_axis_label="Rotor speed (rad/s)",
            y_axis_label="Damped natural frequencies (rad/s)",
        )

        for mark, whirl_dir, legend in zip(
            ["^", "o", "v"], [0.0, 0.5, 1.0], ["Foward", "Mixed", "Backward"]
        ):
            num_frequencies = wd.shape[1]
            for i in range(num_frequencies):
                if wn is True:
                    w_i = wn[:, i]
                else:
                    w_i = wd[:, i]
                whirl_i = whirl[:, i]
                log_dec_i = log_dec[:, i]
                speed_range_i = speed_range[:, i]

                whirl_mask = whirl_i == whirl_dir
                if whirl_mask.shape[0] == 0:
                    continue
                else:
                    im = ax.scatter(
                        speed_range_i[whirl_mask],
                        w_i[whirl_mask],
                        c=log_dec_i[whirl_mask],
                        marker=mark,
                        **kwargs,
                    )

                    # Bokeh plot
                    source = ColumnDataSource(
                        dict(
                            x=speed_range_i[whirl_mask],
                            y=w_i[whirl_mask],
                            color=log_dec_i[whirl_mask],
                        )
                    )
                    color_mapper = linear_cmap(
                        field_name="color",
                        palette=bp.magma(256),
                        low=min(log_dec_map),
                        high=max(log_dec_map),
                    )
                    camp.scatter(
                        x="x",
                        y="y",
                        color=color_mapper,
                        marker=mark,
                        fill_alpha=1.0,
                        size=5,
                        muted_color=color_mapper,
                        muted_alpha=0.2,
                        source=source,
                        legend=legend,
                    )

        ax.plot(
            speed_range[:, 0],
            speed_range[:, 0],
            color="k",
            linewidth=1.5,
            linestyle="-.",
            alpha=0.75,
            label="Rotor speed",
        )

        camp.line(
            x=speed_range[:, 0],
            y=speed_range[:, 0],
            line_width=3,
            color=bokeh_colors[0],
            line_dash="dotdash",
            line_alpha=0.75,
            legend="Rotor speed",
            muted_color=bokeh_colors[0],
            muted_alpha=0.2,
        )

        color_bar = ColorBar(
            color_mapper=color_mapper["transform"], width=8, location=(0, 0)
        )

        camp.legend.click_policy = "mute"
        camp.legend.location = "top_left"
        camp.add_layout(color_bar, "right")
        show(camp)

        if len(fig.axes) == 1:
            cbar = fig.colorbar(im)
            cbar.ax.set_ylabel("log dec")
            cbar.solids.set_edgecolor("face")

            forward_label = mpl.lines.Line2D(
                [], [], marker="^", lw=0, color="tab:blue", alpha=0.3, label="Forward"
            )
            backward_label = mpl.lines.Line2D(
                [], [], marker="v", lw=0, color="tab:blue", alpha=0.3, label="Backward"
            )
            mixed_label = mpl.lines.Line2D(
                [], [], marker="o", lw=0, color="tab:blue", alpha=0.3, label="Mixed"
            )

            legend = plt.legend(
                handles=[forward_label, backward_label, mixed_label], loc=2
            )

            ax.add_artist(legend)

            ax.set_xlabel("Rotor speed ($rad/s$)")
            ax.set_ylabel("Damped natural frequencies ($rad/s$)")

        return fig, ax


class FrequencyResponseResults(Results):
    def plot_magnitude(self, inp, out, ax=None, units="m", **kwargs):
        """Plot frequency response.
        This method plots the frequency response magnitude given an output and
        an input.
        Parameters
        ----------
        inp : int
            Input.
        out : int
            Output.
        ax : matplotlib.axes, optional
            Matplotlib axes where the phase will be plotted.
            If None creates a new.
        kwargs : optional
            Additional key word arguments can be passed to change
            the plot (e.g. linestyle='--')
        Returns
        -------
        ax : matplotlib.axes
            Matplotlib axes with amplitude plot.
        mag_plot : bokeh plot axes
            Bokeh plot axes with amplitude plot.
        Examples
        --------
        """
        if ax is None:
            ax = plt.gca()

        frequency_range = self.frequency_range
        mag = self.magnitude

        ax.plot(frequency_range, mag[inp, out, :], **kwargs)

        ax.set_xlim(0, max(frequency_range))
        ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(prune="lower"))
        ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(prune="upper"))

        if units == "m":
            ax.set_ylabel("Amplitude $(m)$")
            y_axis_label = "Amplitude (m)"
        elif units == "mic-pk-pk":
            ax.set_ylabel("Amplitude $(\mu pk-pk)$")
            y_axis_label = "Amplitude (\mu pk-pk)"
        else:
            ax.set_ylabel("Amplitude $(dB)$")
            y_axis_label = "Amplitude (dB"

        ax.set_xlabel("Frequency (rad/s)")

        # bokeh plot - create a new plot
        mag_plot = figure(
            tools="pan, box_zoom, wheel_zoom, reset, save",
            width=900,
            height=400,
            title="Frequency Response - Magnitude",
            x_axis_label="Frequency",
            y_axis_label=y_axis_label,
        )
        source = ColumnDataSource(dict(x=frequency_range, y=mag[inp, out, :]))
        mag_plot.line(
            x="x",
            y="y",
            source=source,
            line_color=bokeh_colors[0],
            line_alpha=1.0,
            line_width=3,
        )

        return ax, mag_plot

    def plot_phase(self, inp, out, ax=None, **kwargs):
        """Plot frequency response.
        This method plots the frequency response phase given an output and
        an input.
        Parameters
        ----------
        inp : int
            Input.
        out : int
            Output.
        ax : matplotlib.axes, optional
            Matplotlib axes where the phase will be plotted.
            If None creates a new.
        kwargs : optional
            Additional key word arguments can be passed to change
            the plot (e.g. linestyle='--')
        Returns
        -------
        ax : matplotlib.axes
            Matplotlib axes with phase plot.
        phase_plot : bokeh plot axes
            Bokeh plot axes with phase plot.
        Examples
        --------
        """
        if ax is None:
            ax = plt.gca()

        frequency_range = self.frequency_range
        phase = self.phase

        ax.plot(frequency_range, phase[inp, out, :], **kwargs)

        ax.set_xlim(0, max(frequency_range))
        ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(prune="lower"))
        ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(prune="upper"))

        ax.set_ylabel("Phase")
        ax.set_xlabel("Frequency (rad/s)")

        # bokeh plot - create a new plot
        phase_plot = figure(
            tools="pan, box_zoom, wheel_zoom, reset, save",
            width=900,
            height=400,
            title="Frequency Response - Phase",
            x_axis_label="Frequency",
            y_axis_label="Phase",
        )
        source = ColumnDataSource(dict(x=frequency_range, y=phase[inp, out, :]))
        phase_plot.line(
            x="x",
            y="y",
            source=source,
            line_color=bokeh_colors[0],
            line_alpha=1.0,
            line_width=3,
        )

        return ax, phase_plot

    def plot(self, inp, out, ax0=None, ax1=None, **kwargs):
        """Plot frequency response.
        This method plots the frequency response given
        an output and an input.
        Parameters
        ----------
        inp : int
            Input.
        out : int
            Output.
        ax0 : matplotlib.axes, bokeh plot axes optional
            Matplotlib and bokeh plot axes where the amplitude will be plotted.
            If None creates a new.
        ax1 : matplotlib.axes, bokeh plot axes optional
            Matplotlib and bokeh plot axes where the phase will be plotted.
            If None creates a new.
        kwargs : optional
            Additional key word arguments can be passed to change
            the plot (e.g. linestyle='--')
        Returns
        -------
        ax0 : matplotlib.axes
            Matplotlib axes with amplitude plot.
        ax1 : matplotlib.axes
            Matplotlib axes with phase plot.
        bk_ax0 : bokeh plot axes
            Bokeh plot axes with amplitude plot
        bk_ax1 : bokeh plot axes
            Bokeh plot axes with phase plot
        Examples
        --------
        """
        if ax0 is None and ax1 is None:
            fig, (ax0, ax1) = plt.subplots(2)

        # bokeh plot - output to static HTML file
        output_file("plot_freq_response.html")

        # matplotlib axes
        ax0 = self.plot_magnitude(inp, out, ax=ax0)[0]
        ax1 = self.plot_phase(inp, out, ax=ax1)[0]

        # bokeh plot axes
        bk_ax0 = self.plot_magnitude(inp, out, ax=ax0)[1]
        bk_ax1 = self.plot_phase(inp, out, ax=ax1)[1]

        ax0.set_xlabel("")

        # show the bokeh plot results
        grid_plots = gridplot([[bk_ax0], [bk_ax1]])
        show(grid_plots)

        return ax0, ax1, bk_ax0, bk_ax1

    def plot_freq_response_grid(self, outs, inps, ax=None, **kwargs):
        """Plot frequency response.
        This method plots the frequency response given
        an output and an input.
        Parameters
        ----------
        outs : list
            List with the desired outputs.
        inps : list
            List with the desired outputs.
        ax : array with matplotlib.axes, optional
            Matplotlib axes array created with plt.subplots.
            It needs to have a shape of (2*inputs, outputs).
        Returns
        -------
        ax : array with matplotlib.axes, optional
            Matplotlib axes array created with plt.subplots.
        """
        if ax is None:
            fig, ax = plt.subplots(
                len(inps) * 2,
                len(outs),
                sharex=True,
                figsize=(4 * len(outs), 3 * len(inps)),
            )
            fig.subplots_adjust(hspace=0.001, wspace=0.25)

        if len(outs) > 1:
            for i, out in enumerate(outs):
                for j, inp in enumerate(inps):
                    self.plot_magnitude(out, inp, ax=ax[2 * i, j], **kwargs)
                    self.plot_phase(out, inp, ax=ax[2 * i + 1, j], **kwargs)
        else:
            for i, inp in enumerate(inps):
                self.plot_magnitude(outs[0], inp, ax=ax[2 * i], **kwargs)
                self.plot_phase(outs[0], inp, ax=ax[2 * i + 1], **kwargs)

        return ax


class ForcedResponseResults(Results):
    def plot_magnitude(self, dof, ax=None, units="m", **kwargs):
        """Plot frequency response.
        This method plots the frequency response magnitude given an output and
        an input.
        Parameters
        ----------
        dof : int
            Degree of freedom.
        ax : matplotlib.axes, optional
            Matplotlib axes where the phase will be plotted.
            If None creates a new.
        units : str
            Units to plot the magnitude ('m' or 'mic-pk-pk')
        kwargs : optional
            Additional key word arguments can be passed to change
            the plot (e.g. linestyle='--')
        Returns
        -------
        ax : matplotlib.axes
            Matplotlib axes with phase plot.
        mag_plot : bokeh axes
            bokeh axes with magnitude plot
        Examples
        --------
        """
        if ax is None:
            ax = plt.gca()

        frequency_range = self.frequency_range
        mag = self.magnitude

        if units == "m":
            ax.set_ylabel("Amplitude $(m)$")
            y_axis_label = "Amplitude (m)"
        elif units == "mic-pk-pk":
            mag = 2 * mag * 1e6
            ax.set_ylabel("Amplitude $(\mu pk-pk)$")
            y_axis_label = "Amplitude $(\mu pk-pk)$"

        # matplotlib plotting
        ax.plot(frequency_range, mag[dof], **kwargs)

        ax.set_xlim(0, max(frequency_range))
        ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(prune="lower"))
        ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(prune="upper"))

        ax.set_xlabel("Frequency (rad/s)")
        ax.legend()

        # bokeh plot - create a new plot
        mag_plot = figure(
            tools="pan, box_zoom, wheel_zoom, reset, save",
            width=900,
            height=400,
            title="Forced Response - Magnitude",
            x_axis_label="Frequency",
            x_range=[0, max(frequency_range)],
            y_axis_label=y_axis_label,
        )
        source = ColumnDataSource(dict(x=frequency_range, y=mag[dof]))
        mag_plot.line(
            x="x",
            y="y",
            source=source,
            line_color=bokeh_colors[0],
            line_alpha=1.0,
            line_width=3,
        )

        return ax, mag_plot

    def plot_phase(self, dof, ax=None, **kwargs):
        """Plot frequency response.
        This method plots the frequency response phase given an output and
        an input.
        Parameters
        ----------
        dof : int
            Degree of freedom.
        ax : matplotlib.axes, optional
            Matplotlib axes where the phase will be plotted.
            If None creates a new.
        kwargs : optional
            Additional key word arguments can be passed to change
            the plot (e.g. linestyle='--')
        Returns
        -------
        ax : matplotlib.axes
            Matplotlib axes with phase plot.
        phase_plot : bokeh axes
            Bokeh axes with phase plot
        Examples
        --------
        """
        if ax is None:
            ax = plt.gca()

        frequency_range = self.frequency_range
        phase = self.phase

        ax.plot(frequency_range, phase[dof], **kwargs)

        ax.set_xlim(0, max(frequency_range))
        ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(prune="lower"))
        ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(prune="upper"))

        ax.set_ylabel("Phase")
        ax.set_xlabel("Frequency (rad/s)")
        ax.legend()

        # bokeh plot - create a new plot
        phase_plot = figure(
            tools="pan, box_zoom, wheel_zoom, reset, save",
            width=900,
            height=400,
            title="Forced Response - Magnitude",
            x_axis_label="Frequency",
            x_range=[0, max(frequency_range)],
            y_axis_label="Phase",
        )
        source = ColumnDataSource(dict(x=frequency_range, y=phase[dof]))
        phase_plot.line(
            x="x",
            y="y",
            source=source,
            line_color=bokeh_colors[0],
            line_alpha=1.0,
            line_width=3,
        )

        return ax, phase_plot

    def plot(self, dof, ax0=None, ax1=None, **kwargs):
        """Plot frequency response.
        This method plots the frequency response given
        an output and an input.
        Parameters
        ----------
        dof : int
            Degree of freedom.
        ax0 : matplotlib.axes, optional
            Matplotlib axes where the amplitude will be plotted.
            If None creates a new.
        ax1 : matplotlib.axes, optional
            Matplotlib axes where the phase will be plotted.
            If None creates a new.
        kwargs : optional
            Additional key word arguments can be passed to change
            the plot (e.g. linestyle='--')
        Returns
        -------
        ax0 : matplotlib.axes
            Matplotlib axes with amplitude plot.
        ax1 : matplotlib.axes
            Matplotlib axes with phase plot.
        bk_ax0 : bokeh axes
            Bokeh axes with amplitude plot.
        bk_ax1 : bokeh axes
            Bokeh axes with phase plot.
        Examples
        --------
        """
        if ax0 is None and ax1 is None:
            fig, (ax0, ax1) = plt.subplots(2)

        ax0 = self.plot_magnitude(dof, ax=ax0, **kwargs)[0]
        # remove label from phase plot
        kwargs.pop("label", None)
        kwargs.pop("units", None)
        ax1 = self.plot_phase(dof, ax=ax1, **kwargs)[0]

        ax0.set_xlabel("")
        ax0.legend()

        # bokeh plot axes
        bk_ax0 = self.plot_magnitude(dof, ax=ax0, **kwargs)[1]
        bk_ax1 = self.plot_phase(dof, ax=ax1, **kwargs)[1]

        # show the bokeh plot results
        grid_plots = gridplot([[bk_ax0], [bk_ax1]])
        show(grid_plots)

        return ax0, ax1, bk_ax0, bk_ax1


class ModeShapeResults(Results):
    def plot(self, mode=None, evec=None, fig=None, ax=None):
        if ax is None:
            fig = plt.figure()
            ax = fig.gca(projection="3d")

        evec0 = self[:, mode]
        nodes = self.nodes
        nodes_pos = self.nodes_pos
        kappa_modes = self.kappa_modes
        elements_length = self.elements_length

        modex = evec0[0::4]
        modey = evec0[1::4]
        xmax, ixmax = max(abs(modex)), np.argmax(abs(modex))
        ymax, iymax = max(abs(modey)), np.argmax(abs(modey))

        if ymax > 0.4 * xmax:
            evec0 /= modey[iymax]
        else:
            evec0 /= modex[ixmax]

        modex = evec0[0::4]
        modey = evec0[1::4]

        num_points = 201
        c = np.linspace(0, 2 * np.pi, num_points)
        circle = np.exp(1j * c)

        x_circles = np.zeros((num_points, len(nodes)))
        y_circles = np.zeros((num_points, len(nodes)))
        z_circles_pos = np.zeros((num_points, len(nodes)))

        kappa_mode = kappa_modes[mode]

        for node in nodes:
            x = modex[node] * circle
            x_circles[:, node] = np.real(x)
            y = modey[node] * circle
            y_circles[:, node] = np.real(y)
            z_circles_pos[:, node] = nodes_pos[node]

        # plot lines
        nn = 21
        zeta = np.linspace(0, 1, nn)
        onn = np.ones_like(zeta)

        zeta = zeta.reshape(nn, 1)
        onn = onn.reshape(nn, 1)

        xn = np.zeros(nn * (len(nodes) - 1))
        yn = np.zeros(nn * (len(nodes) - 1))
        zn = np.zeros(nn * (len(nodes) - 1))

        N1 = onn - 3 * zeta ** 2 + 2 * zeta ** 3
        N2 = zeta - 2 * zeta ** 2 + zeta ** 3
        N3 = 3 * zeta ** 2 - 2 * zeta ** 3
        N4 = -zeta ** 2 + zeta ** 3

        for Le, n in zip(elements_length, nodes):
            node_pos = nodes_pos[n]
            Nx = np.hstack((N1, Le * N2, N3, Le * N4))
            Ny = np.hstack((N1, -Le * N2, N3, -Le * N4))

            xx = [4 * n, 4 * n + 3, 4 * n + 4, 4 * n + 7]
            yy = [4 * n + 1, 4 * n + 2, 4 * n + 5, 4 * n + 6]

            pos0 = nn * n
            pos1 = nn * (n + 1)
            xn[pos0:pos1] = Nx @ evec0[xx].real
            yn[pos0:pos1] = Ny @ evec0[yy].real
            zn[pos0:pos1] = (node_pos * onn + Le * zeta).reshape(nn)

        for node in nodes:
            ax.plot(
                x_circles[10:, node],
                y_circles[10:, node],
                z_circles_pos[10:, node],
                color=kappa_mode[node],
                linewidth=0.5,
                zdir="x",
            )
            ax.scatter(
                x_circles[10, node],
                y_circles[10, node],
                z_circles_pos[10, node],
                s=5,
                color=kappa_mode[node],
                zdir="x",
            )

        ax.plot(xn, yn, zn, "k--", zdir="x")

        # plot center line
        zn_cl0 = -(zn[-1] * 0.1)
        zn_cl1 = zn[-1] * 1.1
        zn_cl = np.linspace(zn_cl0, zn_cl1, 30)
        ax.plot(zn_cl * 0, zn_cl * 0, zn_cl, "k-.", linewidth=0.8, zdir="x")

        ax.set_zlim(-2, 2)
        ax.set_ylim(-2, 2)
        ax.set_xlim(zn_cl0 - 0.1, zn_cl1 + 0.1)

        ax.set_title(
            f"$speed$ = {self.w:.1f} rad/s\n$"
            f"\frequency_range_d$ = {self.wd[mode]:.1f} rad/s\n"
            f"$log dec$ = {self.log_dec[mode]:.1f}"
        )

        return fig, ax



import bokeh.palettes as bp
import matplotlib.patches as mpatches
import numpy as np
import pytest
import toml
from bokeh.models import ColumnDataSource

from ross.element import Element

__all__ = ["DiskElement"]
bokeh_colors = bp.RdGy[11]


class DiskElement(Element):
    """A disk element.
     This class will create a disk element from input data of inertia and mass.
     Parameters
     ----------
     n: int
         Node in which the disk will be inserted.
     m : float
         Mass of the disk element.
     Id : float
         Diametral moment of inertia.
     Ip : float
         Polar moment of inertia
     References
     ----------
     .. [1] 'Dynamics of Rotating Machinery' by MI Friswell, JET Penny, SD Garvey
        & AW Lees, published by Cambridge University Press, 2010 pp. 156-157.
     Examples
     --------
     >>> disk = DiskElement(0, 32.58972765, 0.17808928, 0.32956362)
     >>> disk.Ip
     0.32956362
     """

    def __init__(self, n, m, Id, Ip):
        self.n = int(n)
        self.n_l = n
        self.n_r = n

        self.m = m
        self.Id = Id
        self.Ip = Ip
        self.color = "#bc625b"

    def __eq__(self, other):
        false_number = 0
        for i in self.__dict__:
            try:
                if pytest.approx(self.__dict__[i]) == other.__dict__[i]:
                    pass
                else:
                    false_number += 1

            except TypeError:
                if self.__dict__[i] == other.__dict__[i]:
                    pass
                else:
                    false_number += 1

        if false_number == 0:
            return True
        else:
            return False

    def __repr__(self):
        return (
            f"{self.__class__.__name__}"
            f"(Id={self.Id:{0}.{5}}, Ip={self.Ip:{0}.{5}}, "
            f"m={self.m:{0}.{5}}, color={self.color!r}, "
            f"n={self.n})"
        )

    def save(self, file_name):
        data = self.load_data(file_name)
        data["DiskElement"][str(self.n)] = {
            "n": self.n,
            "m": self.m,
            "Id": self.Id,
            "Ip": self.Ip,
        }
        self.dump_data(data, file_name)

    @staticmethod
    def load(file_name="DiskElement"):
        disk_elements = []
        with open("DiskElement.toml", "r") as f:
            disk_elements_dict = toml.load(f)
            for element in disk_elements_dict["DiskElement"]:
                disk_elements.append(
                    DiskElement(**disk_elements_dict["DiskElement"][element])
                )
        return disk_elements

    def M(self):
        """
        This method will return the mass matrix for an instance of a disk
        element.
        Parameters
        ----------
        self
        Returns
        -------
        Mass matrix for the disk element.
        Examples
        --------
        >>> disk = DiskElement(0, 32.58972765, 0.17808928, 0.32956362)
        >>> disk.M()
        array([[32.58972765,  0.        ,  0.        ,  0.        ],
               [ 0.        , 32.58972765,  0.        ,  0.        ],
               [ 0.        ,  0.        ,  0.17808928,  0.        ],
               [ 0.        ,  0.        ,  0.        ,  0.17808928]])
        """
        m = self.m
        Id = self.Id
        # fmt: off
        M = np.array([[m, 0,  0,  0],
                      [0, m,  0,  0],
                      [0, 0, Id,  0],
                      [0, 0,  0, Id]])
        # fmt: on
        return M

    def K(self):
        K = np.zeros((4, 4))

        return K

    def C(self):
        C = np.zeros((4, 4))

        return C

    def G(self):
        """
        This method will return the gyroscopic matrix for an instance of a disk
        element.
        Parameters
        ----------
        self
        Returns
        -------
        Gyroscopic matrix for the disk element.
        Examples
        --------
        >>> disk = DiskElement(0, 32.58972765, 0.17808928, 0.32956362)
        >>> disk.G()
        array([[ 0.        ,  0.        ,  0.        ,  0.        ],
               [ 0.        ,  0.        ,  0.        ,  0.        ],
               [ 0.        ,  0.        ,  0.        ,  0.32956362],
               [ 0.        ,  0.        , -0.32956362,  0.        ]])
        """

        Ip = self.Ip
        # fmt: off
        G = np.array([[0, 0,   0,  0],
                      [0, 0,   0,  0],
                      [0, 0,   0, Ip],
                      [0, 0, -Ip,  0]])
        # fmt: on
        return G

    def patch(self, position, length, ax, bk_ax):
        """Disk element patch.
        Patch that will be used to draw the disk element.
        Parameters
        ----------
        ax : matplotlib axes, optional
            Axes in which the plot will be drawn.
        bk_ax : bokeh plotting axes, optional
            Axes in which the plot will be drawn.
        position : float
            Position in which the patch will be drawn.
        length : float
            minimum length of shaft elements
        Returns
        -------
        ax : matplotlib axes
            Returns the axes object with the plot.
        bk_ax : bokeh plotting axes
            Returns the axes object with the plot.
        """
        zpos, ypos = position
        le = length / 8
        D = ypos * 2
        hw = 0.02

        #  matplotlib node (x pos), outer diam. (y pos)
        disk_points_u = [
            [zpos, ypos],  # upper
            [zpos + hw, ypos + D],
            [zpos - hw, ypos + D],
            [zpos, ypos],
        ]
        disk_points_l = [
            [zpos, -ypos],  # lower
            [zpos + hw, -(ypos + D)],
            [zpos - hw, -(ypos + D)],
            [zpos, -ypos],
        ]

        ax.add_patch(mpatches.Polygon(disk_points_u, facecolor=self.color))
        ax.add_patch(mpatches.Polygon(disk_points_l, facecolor=self.color))

        ax.add_patch(mpatches.Circle(xy=(zpos, ypos + D), radius=hw, color=self.color))
        ax.add_patch(
            mpatches.Circle(xy=(zpos, -(ypos + D)), radius=hw, color=self.color)
        )

        # bokeh plot - coordinates to plot disks elements
        z_upper = [zpos, zpos + le, zpos - le]
        y_upper = [ypos, ypos * 4, ypos * 4]

        z_lower = [zpos, zpos + le, zpos - le]
        y_lower = [-ypos, -ypos * 4, -ypos * 4]

        z_circle = z_upper[0]
        y_circle = y_upper[1]

        source = ColumnDataSource(
            dict(z_l=z_lower, y_l=y_lower, z_u=z_upper, y_u=y_upper)
        )

        bk_ax.patch(
            x="z_u",
            y="y_u",
            source=source,
            alpha=1,
            line_width=2,
            color=bokeh_colors[9],
            legend="Disk",
        )
        bk_ax.patch(
            x="z_l",
            y="y_l",
            source=source,
            alpha=1,
            line_width=2,
            color=bokeh_colors[9],
        )
        bk_ax.circle(
            x=z_circle, y=y_circle, radius=le, fill_alpha=1, color=bokeh_colors[9]
        )
        bk_ax.circle(
            x=z_circle, y=-y_circle, radius=le, fill_alpha=1, color=bokeh_colors[9]
        )

    @classmethod
    def from_geometry(cls, n, material, width, i_d, o_d):
        """A disk element.
        This class will create a disk element from input data of geometry.
        Parameters
        ----------
        n: int
            Node in which the disk will be inserted.
        material : lavirot.Material
             Shaft material.
        width: float
            The disk width.
        i_d: float
            Inner diameter.
        o_d: float
            Outer diameter.
        Attributes
        ----------
        m : float
            Mass of the disk element.
        Id : float
            Diametral moment of inertia.
        Ip : float
            Polar moment of inertia
        References
        ----------
        .. [1] 'Dynamics of Rotating Machinery' by MI Friswell, JET Penny, SD Garvey
           & AW Lees, published by Cambridge University Press, 2010 pp. 156-157.
        Examples
        --------
        >>> from ross.materials import steel
        >>> disk = DiskElement.from_geometry(0, steel, 0.07, 0.05, 0.28)
        >>> disk.Ip
        0.32956362089137037
        """
        m = 0.25 * material.rho * np.pi * width * (o_d ** 2 - i_d ** 2)
        # fmt: off
        Id = (
            0.015625 * material.rho * np.pi * width * (o_d ** 4 - i_d ** 4)
            + m * (width ** 2) / 12
        )
        # fmt: on
        Ip = 0.03125 * material.rho * np.pi * width * (o_d ** 4 - i_d ** 4)

        return cls(n, m, Id, Ip)



from abc import ABC, abstractmethod

import pandas as pd
import toml


class Element(ABC):
    """Element class.
    This class is a general class to be called for other files which
    create specific elements for the user
    """

    def __init__(self):
        pass

    def save(self, file_name):
        pass

    @staticmethod
    def load(self, file_name):
        pass

    @abstractmethod
    def M(self):
        pass

    @abstractmethod
    def C(self):
        pass

    @abstractmethod
    def K(self):
        pass

    def summary(self):
        """A summary for the element.

        A pandas series with the element properties as variables.
        """
        attributes = self.__dict__
        attributes["type"] = self.__class__.__name__
        return pd.Series(attributes)

    @staticmethod
    def load_data(file_name):
        try:
            with open(file_name, "r") as f:
                data = toml.load(f)
                if data == {"": {}}:
                    data = {file_name[:-5]: {}}

        except FileNotFoundError:
            data = {file_name[:-5]: {}}
            Element.dump_data(data, file_name)
        return data

    @staticmethod
    def dump_data(data, file_name):
        with open(file_name, "w") as f:
            toml.dump(data, f)

